from __future__ import annotations

from sgnts.base.array_backend import Array as Array
from sgnts.base.array_backend import ArrayBackend as ArrayBackend
from sgnts.base.numpy_backend import NumpyArray as NumpyArray
from sgnts.base.numpy_backend import NumpyBackend as NumpyBackend

try:
    from sgnts.base.torch_backend import TorchArray as TorchArray
    from sgnts.base.torch_backend import TorchBackend as TorchBackend
except ImportError:
    from sgnts.base.notorch_backend import (  # type: ignore[assignment]  # noqa: E501
        TorchArray as TorchArray,
    )
    from sgnts.base.notorch_backend import (  # type: ignore[assignment]  # noqa: E501
        TorchBackend as TorchBackend,
    )
