.md-header {
  background-color: white;
  color: black;
}

.md-grid {
  max-width: 90%;
}

.md-content {
  box-shadow: -0.5px -0.5px 1px 1px gray;
  padding: 1;
}

.md-header__button.md-logo {
  margin-top: 0;
  margin-bottom: 0;
  padding-top: 1;
  padding-bottom: 1;
}

.md-header__button.md-logo img,
.md-header__button.md-logo svg {
  height: 40%;
  width: 40%;
}

